from __future__ import annotations

import inspect
from typing import Any, Dict, Optional, cast, overload

from di._docstrings import join_docstring_from
from di._inspect import DependencyParameter, get_parameters, infer_call_from_annotation
from di.exceptions import WiringError
from di.types.dependencies import DependantProtocol
from di.types.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
    GeneratorProvider,
)
from di.types.scopes import Scope

_VARIABLE_PARAMETER_KINDS = (
    inspect.Parameter.VAR_POSITIONAL,
    inspect.Parameter.VAR_KEYWORD,
)


_expected_attributes = ("call", "scope", "share", "get_dependencies")


def _is_dependant_protocol_instance(o: object) -> bool:
    # run cheap attribute checks before running isinstance
    # isinstance is expensive since it runs reflection on methods
    # to check argument types, etc.
    for attr in _expected_attributes:
        if not hasattr(o, attr):
            return False
    return isinstance(o, DependantProtocol)


class Dependant(DependantProtocol[DependencyType], object):
    @overload
    def __init__(
        self,
        call: Optional[AsyncGeneratorProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        share: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CoroutineProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        share: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[GeneratorProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        share: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CallableProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        share: bool = True,
    ) -> None:
        ...

    def __init__(
        self,
        call: Optional[DependencyProviderType[DependencyType]] = None,
        scope: Scope = None,
        share: bool = True,
    ) -> None:
        self.call = call
        self.scope = scope
        self.dependencies: Optional[
            Dict[str, DependencyParameter[DependantProtocol[Any]]]
        ] = None
        self.share = share

    def __repr__(self) -> str:
        share = "" if self.share is False else ", share=False"
        return f"{self.__class__.__name__}(call={self.call}, scope={self.scope}{share})"

    @join_docstring_from(DependantProtocol[Any].__hash__)
    def __hash__(self) -> int:
        return hash(self.call)

    @join_docstring_from(DependantProtocol[Any].__eq__)
    def __eq__(self, o: object) -> bool:
        if type(self) is not type(o):
            return False
        assert isinstance(o, type(self))
        if self.share is False or o.share is False:
            return False
        return self.call is o.call

    @join_docstring_from(DependantProtocol[Any].get_dependencies)
    def get_dependencies(
        self,
    ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
        """For the Dependant implementation, this serves as a cache layer on
        top of gather_dependencies.
        """
        if self.dependencies is None:  # type: ignore
            self.dependencies = self.gather_dependencies()
        return self.dependencies

    def gather_parameters(self) -> Dict[str, inspect.Parameter]:
        """Collect parameters that this dependency needs to construct itself.

        Generally, this means introspecting into our own callable (self.call).
        """
        assert self.call is not None, "Cannot gather parameters without a bound call"
        return get_parameters(self.call)

    @join_docstring_from(DependantProtocol[Any].infer_call_from_annotation)
    def infer_call_from_annotation(
        self, param: inspect.Parameter
    ) -> DependencyProvider:
        if param.annotation is param.empty:
            raise WiringError(
                "Cannot wire a parameter with no default and no type annotation"
            )
        return infer_call_from_annotation(param)

    def gather_dependencies(
        self,
    ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
        """Collect this dependencies sub dependencies.

        The returned dict corresponds to keyword arguments that will be passed
        to this dependencies `call` after all sub-dependencies are themselves resolved.
        """
        assert (
            self.call is not None
        ), "Container should have assigned call; this is a bug!"
        res: Dict[str, DependencyParameter[DependantProtocol[Any]]] = {}
        for param_name, param in self.gather_parameters().items():
            if param.kind in _VARIABLE_PARAMETER_KINDS:
                raise WiringError(
                    "Dependencies may not use variable positional or keyword arguments"
                )
            if _is_dependant_protocol_instance(param.default):
                sub_dependant = cast(DependantProtocol[Any], param.default)
                if sub_dependant.call is None:
                    sub_dependant.call = sub_dependant.infer_call_from_annotation(param)
            elif param.default is param.empty:
                sub_dependant = self.create_sub_dependant(
                    call=self.infer_call_from_annotation(param),
                    scope=self.scope,
                    share=self.share,
                )
            else:
                continue  # pragma: no cover
            res[param_name] = DependencyParameter(
                dependency=sub_dependant, parameter=param
            )
        return res

    def create_sub_dependant(
        self, call: DependencyProvider, scope: Scope, share: bool
    ) -> DependantProtocol[Any]:
        """Create a Dependant instance from a sub-dependency of this Dependency.

        This is used in the scenario where a transient dependency is inferred from a type hint.
        For example:

        >>> class Foo:
        >>>     ...
        >>> def foo_factory(foo: Foo) -> Foo:
        >>>     return foo
        >>> def parent(foo: Dependency(foo_factory)):
        >>>    ...

        In this scenario, `Dependency(foo_factory)` will call `create_sub_dependant(Foo)`.

        It is recommended to transfer `scope` and possibly `share` to sub-dependencies created in this manner.
        """
        return Dependant[Any](call=call, scope=scope, share=share)



from __future__ import annotations

import functools
from collections import deque
from contextvars import ContextVar
from typing import (
    Any,
    Callable,
    ContextManager,
    Deque,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from di._inspect import (
    DependencyParameter,
    is_async_gen_callable,
    is_coroutine_callable,
)
from di._local_scope_context import LocalScopeContext
from di._state import ContainerState
from di._task import AsyncTask, SyncTask, Task
from di._topsort import topsort
from di.exceptions import (
    DependencyRegistryError,
    DuplicateScopeError,
    ScopeViolationError,
    UnknownScopeError,
)
from di.executors import DefaultExecutor
from di.types import FusedContextManager
from di.types.dependencies import DependantProtocol
from di.types.executor import AsyncExecutor, SyncExecutor
from di.types.providers import (
    Dependency,
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
)
from di.types.scopes import Scope
from di.types.solved import SolvedDependency


class Container:
    def __init__(
        self, executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None
    ) -> None:
        self._context = ContextVar[ContainerState]("context")
        state = ContainerState()
        state.cached_values.add_scope("container")
        state.cached_values.set(Container, self, scope="container")
        self._context.set(state)
        self._executor: Union[AsyncExecutor, SyncExecutor] = (
            executor or DefaultExecutor()
        )

    @property
    def _state(self) -> ContainerState:
        return self._context.get()

    def enter_global_scope(self, scope: Scope) -> FusedContextManager[None]:
        """Enter a global scope that is share amongst threads and coroutines.

        If you enter a global scope in one thread / coroutine, it will propagate to others.
        """
        return self._state.enter_scope(scope)

    def enter_local_scope(self, scope: Scope) -> FusedContextManager[None]:
        """Enter a local scope that is localized to the current thread or coroutine.

        If you enter a global scope in one thread / coroutine, it will NOT propagate to others.
        """
        if scope in self._state.stacks:
            raise DuplicateScopeError(f"Scope {scope} has already been entered!")
        return LocalScopeContext(self._context, scope)

    def bind(
        self,
        provider: DependantProtocol[DependencyType],
        dependency: DependencyProviderType[DependencyType],
    ) -> ContextManager[None]:
        """Bind a new dependency provider for a given dependency.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).

        Binds are only identified by the identity of the callable and do not take into account
        the scope or any other data from the dependency they are replacing.

        The `scope` parameter determines the scope for the bind itself.
        The bind will be automatically cleared when that scope is exited.
        If no scope is provided, the current scope is used.
        """
        return self._state.bind(provider=provider, dependency=dependency)

    def solve(
        self, dependency: DependantProtocol[DependencyType]
    ) -> SolvedDependency[DependencyType]:
        """Solve a dependency.

        This is done automatically when calling `execute`, but you can store the returned value
        from this function and call `execute_solved` instead if you know that your binds
        will not be changing between calls.
        """

        if dependency.call in self._state.binds:
            dependency = self._state.binds[dependency.call]  # type: ignore

        param_graph: Dict[
            DependantProtocol[Any],
            Dict[str, DependencyParameter[DependantProtocol[Any]]],
        ] = {}

        dep_registry: Dict[DependantProtocol[Any], DependantProtocol[Any]] = {}

        dep_dag: Dict[DependantProtocol[Any], List[DependantProtocol[Any]]] = {}

        def get_params(
            dep: DependantProtocol[Any],
        ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
            params = dep.get_dependencies().copy()
            for keyword, param in params.items():
                assert param.dependency.call is not None
                if param.dependency.call in self._state.binds:
                    params[keyword] = DependencyParameter[Any](
                        dependency=self._state.binds.get(param.dependency.call),
                        parameter=param.parameter,
                    )
            return params

        def check_equivalent(dep: DependantProtocol[Any]):
            if dep in dep_registry and dep.scope != dep_registry[dep].scope:
                raise DependencyRegistryError(
                    f"The dependencies {dep} and {dep_registry[dep]}"
                    " have the same lookup (__hash__ and __eq__) but have different scopes"
                    f" ({dep.scope} and {dep_registry[dep].scope} respectively)"
                    " This is likely a mistake, but you can override this behavior by either:"
                    "\n  1. Wrapping the function in another function or subclassing the class such"
                    " that they now are considered different dependencies"
                    "\n  2. Use a custom implemetnation of DependantProtocol that changes the meaning"
                    " or computation of __hash__ and/or __eq__"
                )

        q: Deque[DependantProtocol[Any]] = deque([dependency])
        while q:
            dep = q.popleft()
            if dep in dep_registry:
                check_equivalent(dep)
            else:
                dep_registry[dep] = dep
                params = get_params(dep)
                param_graph[dep] = params
                dep_dag[dep] = []
                for param in params.values():
                    subdep = param.dependency
                    dep_dag[dep].append(subdep)
                    if subdep not in dep_registry:
                        q.append(subdep)

        topsorted_groups = topsort(dependency, dep_dag)
        tasks, get_results = self._build_tasks(
            topsorted_groups, dependency, param_graph
        )
        return SolvedDependency(
            dependency=dependency,
            dag=param_graph,
            _tasks=tasks,
            _get_results=get_results,
        )

    def _build_task(
        self,
        dependency: DependantProtocol[DependencyType],
        tasks: Dict[
            DependantProtocol[Any], Union[AsyncTask[Dependency], SyncTask[Dependency]]
        ],
        dag: Dict[
            DependantProtocol[Any],
            Dict[str, DependencyParameter[DependantProtocol[Any]]],
        ],
    ) -> Union[AsyncTask[DependencyType], SyncTask[DependencyType]]:

        task_dependencies: Dict[str, DependencyParameter[Task[DependencyProvider]]] = {}

        for param_name, param in dag[dependency].items():
            task_dependencies[param_name] = DependencyParameter(
                dependency=tasks[param.dependency], parameter=param.parameter
            )

        if is_async_gen_callable(dependency.call) or is_coroutine_callable(
            dependency.call
        ):
            return AsyncTask(dependant=dependency, dependencies=task_dependencies)
        else:
            return SyncTask(dependant=dependency, dependencies=task_dependencies)

    def _build_tasks(
        self,
        topsort: List[List[DependantProtocol[Any]]],
        dependency: DependantProtocol[DependencyType],
        dag: Dict[
            DependantProtocol[Any],
            Dict[str, DependencyParameter[DependantProtocol[Any]]],
        ],
    ) -> Tuple[
        List[List[Union[AsyncTask[Dependency], SyncTask[Dependency]]]],
        Callable[[], DependencyType],
    ]:
        tasks: Dict[
            DependantProtocol[Any], Union[AsyncTask[Dependency], SyncTask[Dependency]]
        ] = {}
        for group in reversed(topsort):
            for dep in group:
                if dep not in tasks:
                    tasks[dep] = self._build_task(dep, tasks, dag)
        get_result = tasks[dependency].get_result
        return (
            list(reversed([[tasks[dep] for dep in group] for group in topsort])),
            get_result,
        )

    def _validate_scopes(self, solved: SolvedDependency[Dependency]) -> None:
        """Validate that dependencies all have a valid scope and
        that dependencies only depend on outer scopes or their own scope.
        """
        scopes: Dict[Scope, int] = {
            scope: idx
            for idx, scope in enumerate(reversed(self._state.scopes + [None]))
        }

        def check_is_inner(
            dep: DependantProtocol[Any], subdep: DependantProtocol[Any]
        ) -> None:
            if scopes[dep.scope] > scopes[subdep.scope]:
                raise ScopeViolationError(
                    f"{dep} cannot depend on {subdep} because {subdep}'s"
                    f" scope ({subdep.scope}) is narrower than {dep}'s scope ({dep.scope})"
                )

        def check_scope(dep: DependantProtocol[Any]) -> None:
            if dep.scope not in scopes:
                raise UnknownScopeError(
                    f"Dependency{dep} has an unknown scope {dep.scope}."
                    f" Did you forget to enter the {dep.scope} scope?"
                )

        for dep, params in solved.dag.items():
            check_scope(dep)
            for param in params.values():
                subdep = param.dependency
                check_scope(subdep)
                check_is_inner(dep, subdep)

    def execute_sync(
        self,
        solved: SolvedDependency[DependencyType],
        validate_scopes: bool = True,
    ) -> DependencyType:
        """Execute an already solved dependency.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        with self.enter_local_scope(None):
            if validate_scopes:
                self._validate_scopes(solved)

            tasks, get_results = solved._tasks, solved._get_results  # type: ignore

            if not hasattr(self._executor, "execute_sync"):
                raise TypeError(
                    "execute_sync requires an executor implementing the SyncExecutor protocol"
                )
            executor = cast(SyncExecutor, self._executor)

            stateful_tasks = [
                [functools.partial(t.compute, self._state) for t in group]
                for group in tasks
            ]

            return executor.execute_sync(stateful_tasks, get_results)  # type: ignore

    async def execute_async(
        self,
        solved: SolvedDependency[DependencyType],
        validate_scopes: bool = True,
    ) -> DependencyType:
        """Execute an already solved dependency.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        async with self.enter_local_scope(None):
            if validate_scopes:
                self._validate_scopes(solved)

            tasks, get_results = solved._tasks, solved._get_results  # type: ignore

            if not hasattr(self._executor, "execute_async"):
                raise TypeError(
                    "execute_async requires an executor implementing the AsyncExecutor protocol"
                )
            executor = cast(AsyncExecutor, self._executor)

            stateful_tasks = [
                [functools.partial(t.compute, self._state) for t in group]
                for group in tasks
            ]

            return await executor.execute_async(stateful_tasks, get_results)  # type: ignore



from __future__ import annotations

import inspect
from typing import Any, Dict, Optional, Protocol, runtime_checkable

from di._inspect import DependencyParameter
from di.types.providers import (
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
)
from di.types.scopes import Scope


@runtime_checkable
class DependantProtocol(Protocol[DependencyType]):
    """A dependant is an object that can provide the container with:
    - A hash, to compare itself against other dependants
    - A scope
    - A callable that can be used to assemble itself
    - The dependants that correspond to the keyword arguments of that callable
    """

    call: Optional[DependencyProviderType[DependencyType]]
    dependencies: Optional[Dict[str, DependencyParameter[DependantProtocol[Any]]]]
    scope: Scope
    share: bool

    def __hash__(self) -> int:
        """Dependencies need to be hashable for quick O(1) lookups.

        Generally, hash(self.call) will suffice so that dependencies are identified by their callable.
        """
        raise NotImplementedError

    def __eq__(self, o: object) -> bool:
        """Used in conjunction with __hash__ for mapping lookups of dependencies.

        If this returns `True`, the two dependencies are considered the same and
        `di` will pick one to subsitute for the other.

        Note that using the same dependenyc in two different scopes is prohibited,
        so if this returns `True` and `self.scope != o.scope` `di` will raise a DependencyRegistryError.
        """
        raise NotImplementedError

    def get_dependencies(
        self,
    ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
        """Collect all of the sub dependencies for this dependant into a mapping
        of parameter name => parameter spec.
        """
        raise NotImplementedError

    def infer_call_from_annotation(
        self, param: inspect.Parameter
    ) -> DependencyProvider:
        """Called when the dependency was not explicitly passed a callable.

        It is important to note that param in this context refers to the parameter in this
        Dependant's parent.
        For example, in the case of `def func(thing: Something = Dependency())` this method
        will be called with a Parameter corresponding to Something.
        """
        raise NotImplementedError



from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import TYPE_CHECKING, Dict, Generic, List, NewType, Tuple, Union, cast

from di._inspect import DependencyParameter, is_async_gen_callable, is_gen_callable
from di._state import ContainerState
from di.exceptions import IncompatibleDependencyError
from di.types.dependencies import DependantProtocol
from di.types.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    Dependency,
    DependencyType,
    GeneratorProvider,
)

_UNSET_TYPE = NewType("_UNSET_TYPE", object)
_UNSET = _UNSET_TYPE(object())


class Task(Generic[DependencyType]):
    dependant: DependantProtocol[DependencyType]
    dependencies: Dict[str, DependencyParameter[Task[Dependency]]]
    _result: Union[_UNSET_TYPE, DependencyType]

    __slots__ = ("dependant", "dependencies", "_result")

    def __init__(
        self,
        dependant: DependantProtocol[DependencyType],
        dependencies: Dict[str, DependencyParameter[Task[Dependency]]],
    ) -> None:
        self.dependant = dependant
        self.dependencies = dependencies
        self._result = _UNSET

    def _gather_params(self) -> Tuple[List[Dependency], Dict[str, Dependency]]:
        positional: List[Dependency] = []
        keyword: Dict[str, Dependency] = {}
        for param_name, dep in self.dependencies.items():
            if dep.parameter.kind is dep.parameter.kind.POSITIONAL_ONLY:
                positional.append(dep.dependency.get_result())
            else:
                keyword[param_name] = dep.dependency.get_result()
        return positional, keyword

    def get_result(self) -> DependencyType:
        if self._result is _UNSET:
            raise ValueError(
                "`compute()` must be called before `get_result()`; this is likely a bug!"
            )
        return cast(DependencyType, self._result)


class AsyncTask(Task[DependencyType]):
    async def compute(self, state: ContainerState) -> None:
        assert self.dependant.call is not None
        call = self.dependant.call
        args, kwargs = self._gather_params()

        if self.dependant.share and state.cached_values.contains(self.dependant.call):
            # use cached value
            self._result = state.cached_values.get(self.dependant.call)
        else:
            if is_async_gen_callable(self.dependant.call):
                stack = state.stacks[self.dependant.scope]
                if not isinstance(stack, AsyncExitStack):
                    raise IncompatibleDependencyError(
                        f"The dependency {self.dependant} is an awaitable dependency"
                        f" and canot be used in the sync scope {self.dependant.scope}"
                    )
                if TYPE_CHECKING:
                    call = cast(AsyncGeneratorProvider[DependencyType], call)
                self._result = await stack.enter_async_context(
                    asynccontextmanager(call)(*args, **kwargs)
                )
            else:
                if TYPE_CHECKING:
                    call = cast(CoroutineProvider[DependencyType], call)
                self._result = await call(*args, **kwargs)
            if self.dependant.share:
                # caching is allowed, now that we have a value we can save it and start using the cache
                state.cached_values.set(
                    self.dependant.call, self._result, scope=self.dependant.scope
                )


class SyncTask(Task[DependencyType]):
    def compute(self, state: ContainerState) -> None:
        assert self.dependant.call is not None
        call = self.dependant.call
        args, kwargs = self._gather_params()

        if self.dependant.share and state.cached_values.contains(self.dependant.call):
            # use cached value
            self._result = state.cached_values.get(self.dependant.call)
        else:
            if is_gen_callable(self.dependant.call):
                if TYPE_CHECKING:
                    call = cast(GeneratorProvider[DependencyType], call)
                stack = state.stacks[self.dependant.scope]
                self._result = stack.enter_context(
                    contextmanager(call)(*args, **kwargs)
                )
            else:
                if TYPE_CHECKING:
                    call = cast(CallableProvider[DependencyType], call)
                self._result = call(*args, **kwargs)
            if self.dependant.share:
                # caching is allowed, now that we have a value we can save it and start using the cache
                state.cached_values.set(
                    self.dependant.call, self._result, scope=self.dependant.scope
                )



from __future__ import annotations

import dataclasses
import functools
from typing import Any, Callable, Dict, Generic, List, Union

from di._inspect import DependencyParameter
from di._task import AsyncTask, SyncTask
from di.types.dependencies import DependantProtocol
from di.types.providers import Dependency, DependencyType


@dataclasses.dataclass(frozen=True)
class SolvedDependency(Generic[DependencyType]):
    """Representation of a fully solved dependency.

    A fully solved dependency consists of:
    - A DAG of sub-dependency paramters.
    - A topologically sorted order of execution, where each sublist represents a
    group of dependencies that can be executed in parallel.
    """

    dependency: DependantProtocol[DependencyType]
    dag: Dict[
        DependantProtocol[Any], Dict[str, DependencyParameter[DependantProtocol[Any]]]
    ]
    _tasks: List[List[Union[AsyncTask[Dependency], SyncTask[Dependency]]]]
    _get_results: Callable[[], DependencyType]

    @functools.cached_property
    def flat_subdependants(self) -> List[DependantProtocol[Any]]:
        """Get an exhaustive list of all of the dependencies of this dependency,
        in no particular order.
        """
        return list(self.dag.keys() - {self.dependency})



import concurrent.futures
import functools
import inspect
import typing

import anyio
import anyio.abc

from di._concurrency import curry_context, gurantee_awaitable
from di.types.executor import AsyncExecutor, SyncExecutor, Task

ResultType = typing.TypeVar("ResultType")
T = typing.TypeVar("T")


def _all_sync(tasks: typing.List[Task]) -> bool:
    for task in tasks:
        if isinstance(task, functools.partial):
            call = task.func  # type: ignore
        else:
            call = task  # pragma: no cover
        if inspect.iscoroutinefunction(call):
            return False
    return True


class DefaultExecutor(AsyncExecutor, SyncExecutor):
    def __init__(self) -> None:
        self._threadpool = concurrent.futures.ThreadPoolExecutor()

    def execute_sync(
        self,
        tasks: typing.List[typing.List[Task]],
        get_result: typing.Callable[[], ResultType],
    ) -> ResultType:
        for task_group in tasks:
            if len(task_group) > 1:
                futures: typing.List[
                    concurrent.futures.Future[
                        typing.Union[None, typing.Awaitable[None]]
                    ]
                ] = []
                for task in task_group:
                    futures.append(self._threadpool.submit(curry_context(task)))
                for future in concurrent.futures.as_completed(futures):
                    exc = future.exception()
                    if exc is not None:
                        raise exc
                    if inspect.isawaitable(future.result()):
                        raise TypeError(
                            "Cannot execute async dependencies in a SyncExecutor"
                        )
            else:
                v = task_group[0]()
                if inspect.isawaitable(v):
                    raise TypeError(
                        "Cannot execute async dependencies in a SyncExecutor"
                    )
        return get_result()

    async def execute_async(
        self,
        tasks: typing.List[typing.List[Task]],
        get_result: typing.Callable[[], ResultType],
    ) -> ResultType:
        tg: typing.Optional[anyio.abc.TaskGroup] = None
        for task_group in tasks:
            if len(task_group) > 1:
                if _all_sync(task_group):
                    self.execute_sync([task_group], lambda: None)
                else:
                    if tg is None:
                        tg = anyio.create_task_group()
                    async with tg:
                        for task in task_group:
                            tg.start_soon(gurantee_awaitable(task))  # type: ignore
            else:
                maybe_awaitable = next(iter(task_group))()
                try:
                    # this could break if a dependency *value* is an awaitable
                    # the day that happens, we can go back to using inspect
                    # otherwise, this is much faster
                    await maybe_awaitable  # type: ignore
                except TypeError:
                    pass  # not awaitable
        return get_result()



"""Convenience functions, mainly for the purpose of providing proper type annotations for default arguments.
"""

from typing import Optional, overload

from di.dependant import Dependant
from di.types.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProviderType,
    DependencyType,
    GeneratorProvider,
)
from di.types.scopes import Scope


@overload
def Depends(
    call: Optional[AsyncGeneratorProvider[DependencyType]] = None,
    *,
    scope: Optional[Scope] = None,
    share: bool = True
) -> DependencyType:
    ...


@overload
def Depends(
    call: Optional[CoroutineProvider[DependencyType]] = None,
    *,
    scope: Optional[Scope] = None,
    share: bool = True
) -> DependencyType:
    ...


@overload
def Depends(
    call: Optional[GeneratorProvider[DependencyType]] = None,
    *,
    scope: Optional[Scope] = None,
    share: bool = True
) -> DependencyType:
    ...


@overload
def Depends(
    call: Optional[CallableProvider[DependencyType]] = None,
    *,
    scope: Optional[Scope] = None,
    share: bool = True
) -> DependencyType:
    ...


def Depends(
    call: Optional[DependencyProviderType[DependencyType]] = None,
    *,
    scope: Optional[Scope] = None,
    share: bool = True
) -> DependencyType:
    return Dependant(call=call, scope=scope, share=share)  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['di', 'di.types']

package_data = \
{'': ['*']}

install_requires = \
['anyio>=3,<4']

setup_kwargs = {
    'name': 'di',
    'version': '0.4.17',
    'description': 'Autowiring dependency injection',
    'long_description': '# di: pythonic dependency injection\n\n[![codecov](https://codecov.io/gh/adriangb/di/branch/main/graph/badge.svg?token=A0FXC8B93Y)](https://codecov.io/gh/adriangb/di)\n![Test & Release](https://github.com/adriangb/di/actions/workflows/release.yaml/badge.svg)\n\n⚠️ This project is a WIP. Until there is a 1.X.Y release, expect breaking changes. ⚠️\n\nFor more information, see our [docs].\n\n[docs]: https://www.adriangb.com/di/\n\nSee this release on GitHub: [v0.4.17](https://github.com/adriangb/di/releases/tag/0.4.17)\n',
    'author': 'Adrian Garcia Badaracco',
    'author_email': 'adrian@adriangb.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/adriangb/di',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4',
}


setup(**setup_kwargs)
